@@ -20,8 +20,8 @@
 #include <cstdint>
 #include <queue>
 #include <vector>
-#include <xmmintrin.h>
 
+#include "./hedly.h"
 #include "./libdivide.h"
 
 #ifndef DIJKSTRA3D_HPP
@@ -59,7 +59,7 @@ inline std::vector<OUT> query_shortest_path(const OUT* parents, const OUT target
   return path;
 }
 
-inline void compute_neighborhood(
+inline void compute_neighborhood_helper(
   int *neighborhood, 
   const int x, const int y, const int z,
   const uint64_t sx, const uint64_t sy, const uint64_t sz,
@@ -108,6 +108,66 @@ inline void compute_neighborhood(
   neighborhood[25] = (connectivity > 18) * (neighborhood[1] + neighborhood[3] + neighborhood[5]) * (neighborhood[3] && neighborhood[5]);
 }
 
+inline void compute_neighborhood(
+  int *neighborhood, 
+  const int x, const int y, const int z,
+  const uint64_t sx, const uint64_t sy, const uint64_t sz,
+  const int connectivity = 26, const uint32_t* voxel_connectivity_graph = NULL) {
+
+  compute_neighborhood_helper(neighborhood, x, y, z, sx, sy, sz, connectivity);
+
+  if (voxel_connectivity_graph == NULL) {
+    return;
+  }
+
+  uint64_t loc = x + sx * (y + sy * z);
+  uint32_t graph = voxel_connectivity_graph[loc];
+
+  // graph conventions are defined here:
+  // https://github.com/seung-lab/connected-components-3d/blob/3.2.0/cc3d_graphs.hpp#L73-L92
+
+  // 6-hood
+  neighborhood[0] *= ((graph & 0b000010) > 0); // -x
+  neighborhood[1] *= ((graph & 0b000001) > 0); // +x
+  neighborhood[2] *= ((graph & 0b001000) > 0); // -y
+  neighborhood[3] *= ((graph & 0b000100) > 0); // +y
+  neighborhood[4] *= ((graph & 0b100000) > 0); // -z
+  neighborhood[5] *= ((graph & 0b010000) > 0); // +z
+
+  // 18-hood
+
+  // xy diagonals
+  neighborhood[6] *= ((graph & 0b1000000000) > 0); // up-left
+  neighborhood[7] *= ((graph & 0b0100000000) > 0); // up-right
+  neighborhood[8] *= ((graph & 0b0010000000) > 0); // down-left
+  neighborhood[9] *= ((graph & 0b0001000000) > 0); // down-right
+
+  // yz diagonals
+  neighborhood[10] *= ((graph & 0b100000000000000000) > 0); // up-left
+  neighborhood[11] *= ((graph & 0b010000000000000000) > 0); // up-right
+  neighborhood[12] *= ((graph & 0b000010000000000000) > 0); // down-left
+  neighborhood[13] *= ((graph & 0b000001000000000000) > 0); // down-right
+
+  // xz diagonals
+  neighborhood[14] *= ((graph & 0b001000000000000000) > 0); // up-left
+  neighborhood[15] *= ((graph & 0b000100000000000000) > 0); // up-right
+  neighborhood[16] *= ((graph & 0b000000100000000000) > 0); // down-left
+  neighborhood[17] *= ((graph & 0b000000010000000000) > 0); // down-right
+
+  // 26-hood
+
+  // Now the eight corners of the cube
+  neighborhood[18] *= ((graph & 0b10000000000000000000000000) > 0);
+  neighborhood[19] *= ((graph & 0b01000000000000000000000000) > 0);
+  neighborhood[20] *= ((graph & 0b00100000000000000000000000) > 0);
+  neighborhood[21] *= ((graph & 0b00001000000000000000000000) > 0);
+  neighborhood[22] *= ((graph & 0b00010000000000000000000000) > 0);
+  neighborhood[23] *= ((graph & 0b00000100000000000000000000) > 0);
+  neighborhood[24] *= ((graph & 0b00000010000000000000000000) > 0);
+  neighborhood[25] *= ((graph & 0b00000001000000000000000000) > 0);
+}
+
+
 template <typename T = uint32_t>
 class HeapNode {
 public:
@@ -137,16 +197,16 @@ struct HeapNodeCompare {
   }
 };
 
-#define DIJKSTRA_3D_PREFETCH_26WAY(field) \
-  _mm_prefetch(reinterpret_cast<char*>(&field[loc - 1]), _MM_HINT_T0); \
-  _mm_prefetch(reinterpret_cast<char*>(&field[loc + sxy - 1]), _MM_HINT_T0); \
-  _mm_prefetch(reinterpret_cast<char*>(&field[loc - sxy - 1]), _MM_HINT_T0); \
-  _mm_prefetch(reinterpret_cast<char*>(&field[loc + sxy + sx - 1]), _MM_HINT_T0); \
-  _mm_prefetch(reinterpret_cast<char*>(&field[loc + sxy - sx - 1]), _MM_HINT_T0); \
-  _mm_prefetch(reinterpret_cast<char*>(&field[loc - sxy + sx - 1]), _MM_HINT_T0); \
-  _mm_prefetch(reinterpret_cast<char*>(&field[loc - sxy - sx - 1]), _MM_HINT_T0); \
-  _mm_prefetch(reinterpret_cast<char*>(&field[loc + sx - 1]), _MM_HINT_T0); \
-  _mm_prefetch(reinterpret_cast<char*>(&field[loc - sx - 1]), _MM_HINT_T0);
+#define DIJKSTRA_3D_PREFETCH_26WAY(field, loc) \
+  HEDLEYX_PREFETCH(reinterpret_cast<char*>(&field[(loc) - 1]), 0, 1); \
+  HEDLEYX_PREFETCH(reinterpret_cast<char*>(&field[(loc) + sxy - 1]), 0, 1); \
+  HEDLEYX_PREFETCH(reinterpret_cast<char*>(&field[(loc) - sxy - 1]), 0, 1); \
+  HEDLEYX_PREFETCH(reinterpret_cast<char*>(&field[(loc) + sxy + sx - 1]), 0, 1); \
+  HEDLEYX_PREFETCH(reinterpret_cast<char*>(&field[(loc) + sxy - sx - 1]), 0, 1); \
+  HEDLEYX_PREFETCH(reinterpret_cast<char*>(&field[(loc) - sxy + sx - 1]), 0, 1); \
+  HEDLEYX_PREFETCH(reinterpret_cast<char*>(&field[(loc) - sxy - sx - 1]), 0, 1); \
+  HEDLEYX_PREFETCH(reinterpret_cast<char*>(&field[(loc) + sx - 1]), 0, 1); \
+  HEDLEYX_PREFETCH(reinterpret_cast<char*>(&field[(loc) - sx - 1]), 0, 1);
 
 /* Perform dijkstra's shortest path algorithm
  * on a 3D image grid. Vertices are voxels and
@@ -177,7 +237,8 @@ std::vector<OUT> dijkstra3d(
     T* field, 
     const size_t sx, const size_t sy, const size_t sz, 
     const size_t source, const size_t target,
-    const int connectivity = 26
+    const int connectivity = 26, 
+    const uint32_t* voxel_connectivity_graph = NULL
   ) {
 
   connectivity_check(connectivity);
@@ -223,9 +284,9 @@ std::vector<OUT> dijkstra3d(
     // As early as possible, start fetching the
     // data from RAM b/c the annotated lines below
     // have 30-50% cache miss.
-    DIJKSTRA_3D_PREFETCH_26WAY(field)
-    DIJKSTRA_3D_PREFETCH_26WAY(dist)
-    
+    DIJKSTRA_3D_PREFETCH_26WAY(field, loc)
+    DIJKSTRA_3D_PREFETCH_26WAY(dist, loc)
+
     if (power_of_two) {
       z = loc >> (xshift + yshift);
       y = (loc - (z << (xshift + yshift))) >> xshift;
@@ -237,7 +298,7 @@ std::vector<OUT> dijkstra3d(
       x = loc - sx * (y + z * sy);
     }
 
-    compute_neighborhood(neighborhood, x, y, z, sx, sy, sz, connectivity);
+    compute_neighborhood(neighborhood, x, y, z, sx, sy, sz, connectivity, voxel_connectivity_graph);
 
     for (int i = 0; i < connectivity; i++) {
       if (neighborhood[i] == 0) {
@@ -314,7 +375,8 @@ std::vector<OUT> bidirectional_dijkstra3d(
     T* field, 
     const size_t sx, const size_t sy, const size_t sz, 
     const size_t source, const size_t target,
-    const int connectivity = 26
+    const int connectivity = 26,
+    const uint32_t* voxel_connectivity_graph = NULL
   ) {
 
   connectivity_check(connectivity);
@@ -401,13 +463,13 @@ std::vector<OUT> bidirectional_dijkstra3d(
       }
     }
 
-    DIJKSTRA_3D_PREFETCH_26WAY(field)
+    DIJKSTRA_3D_PREFETCH_26WAY(field, loc)
 
     if (forward) {
-      DIJKSTRA_3D_PREFETCH_26WAY(dist_fwd)
+      DIJKSTRA_3D_PREFETCH_26WAY(dist_fwd, loc)
     }
     else {
-      DIJKSTRA_3D_PREFETCH_26WAY(dist_rev)
+      DIJKSTRA_3D_PREFETCH_26WAY(dist_rev, loc)
     }
 
     if (power_of_two) {
@@ -421,7 +483,7 @@ std::vector<OUT> bidirectional_dijkstra3d(
       x = loc - sx * (y + z * sy);
     }
 
-    compute_neighborhood(neighborhood, x, y, z, sx, sy, sz, connectivity);
+    compute_neighborhood(neighborhood, x, y, z, sx, sy, sz, connectivity, voxel_connectivity_graph);
 
     if (forward) {
       bidirectional_core<T,OUT>(
@@ -468,7 +530,8 @@ std::vector<OUT> compass_guided_dijkstra3d(
     T* field, 
     const size_t sx, const size_t sy, const size_t sz, 
     const size_t source, const size_t target,
-    const int connectivity = 26, float normalizer = -1
+    const int connectivity = 26, float normalizer = -1,
+    const uint32_t* voxel_connectivity_graph = NULL
   ) {
 
   connectivity_check(connectivity);
@@ -541,12 +604,12 @@ std::vector<OUT> compass_guided_dijkstra3d(
       continue;
     }
 
-    DIJKSTRA_3D_PREFETCH_26WAY(field)
-    DIJKSTRA_3D_PREFETCH_26WAY(dist)
+    DIJKSTRA_3D_PREFETCH_26WAY(field, loc)
+    DIJKSTRA_3D_PREFETCH_26WAY(dist, loc)
 
     xyzfn(loc);
 
-    compute_neighborhood(neighborhood, x, y, z, sx, sy, sz);
+    compute_neighborhood(neighborhood, x, y, z, sx, sy, sz, connectivity, voxel_connectivity_graph);
 
     for (int i = 0; i < connectivity; i++) {
       if (neighborhood[i] == 0) {
@@ -636,7 +699,8 @@ OUT* parental_field3d(
     T* field, 
     const size_t sx, const size_t sy, const size_t sz, 
     const size_t source, OUT* parents = NULL,
-    const int connectivity = 26
+    const int connectivity = 26,
+    const uint32_t* voxel_connectivity_graph = NULL
   ) {
 
   connectivity_check(connectivity);
@@ -679,8 +743,8 @@ OUT* parental_field3d(
       continue;
     }
 
-    DIJKSTRA_3D_PREFETCH_26WAY(field)
-    DIJKSTRA_3D_PREFETCH_26WAY(dist)
+    DIJKSTRA_3D_PREFETCH_26WAY(field, loc)
+    DIJKSTRA_3D_PREFETCH_26WAY(dist, loc)
 
     if (power_of_two) {
       z = loc >> (xshift + yshift);
@@ -693,7 +757,7 @@ OUT* parental_field3d(
       x = loc - sx * (y + z * sy);
     }
 
-    compute_neighborhood(neighborhood, x, y, z, sx, sy, sz, connectivity);
+    compute_neighborhood(neighborhood, x, y, z, sx, sy, sz, connectivity, voxel_connectivity_graph);
 
     for (int i = 0; i < connectivity; i++) {
       if (neighborhood[i] == 0) {
@@ -726,9 +790,12 @@ template <typename T>
 float* distance_field3d(
     T* field, 
     const size_t sx, const size_t sy, const size_t sz, 
-    const size_t source
+    const size_t source, const size_t connectivity=26,
+    const uint32_t* voxel_connectivity_graph = NULL
   ) {
 
+  connectivity_check(connectivity);
+
   const size_t voxels = sx * sy * sz;
   const size_t sxy = sx * sy;
 
@@ -748,7 +815,7 @@ float* distance_field3d(
   std::priority_queue<HeapNode<size_t>, std::vector<HeapNode<size_t>>, HeapNodeCompare<size_t>> queue;
   queue.emplace(0.0, source);
 
-  size_t loc;
+  size_t loc, next_loc;
   float delta;
   size_t neighboridx;
 
@@ -762,6 +829,18 @@ float* distance_field3d(
       continue;
     }
 
+    if (!queue.empty()) {
+      next_loc = queue.top().value;
+      if (!std::signbit(dist[next_loc])) {
+
+        // As early as possible, start fetching the
+        // data from RAM b/c the annotated lines below
+        // have 30-50% cache miss.
+        DIJKSTRA_3D_PREFETCH_26WAY(field, next_loc)
+        DIJKSTRA_3D_PREFETCH_26WAY(dist, next_loc)
+      }
+    }
+
     if (power_of_two) {
       z = loc >> (xshift + yshift);
       y = (loc - (z << (xshift + yshift))) >> xshift;
@@ -773,9 +852,9 @@ float* distance_field3d(
       x = loc - sx * (y + z * sy);
     }
 
-    compute_neighborhood(neighborhood, x, y, z, sx, sy, sz);
+    compute_neighborhood(neighborhood, x, y, z, sx, sy, sz, 26, voxel_connectivity_graph);
 
-    for (int i = 0; i < NHOOD_SIZE; i++) {
+    for (size_t i = 0; i < connectivity; i++) {
       if (neighborhood[i] == 0) {
         continue;
       }
@@ -930,7 +1009,8 @@ float* euclidean_distance_field3d(
     const float wx, const float wy, const float wz, 
     const size_t source, 
     const float free_space_radius = 0,
-    float* dist = NULL
+    float* dist = NULL,
+    const uint32_t* voxel_connectivity_graph = NULL
   ) {
 
   const size_t voxels = sx * sy * sz;
@@ -977,7 +1057,7 @@ float* euclidean_distance_field3d(
     );
   }
 
-  size_t loc;
+  size_t loc, next_loc;
   float new_dist;
   size_t neighboridx;
 
@@ -991,11 +1071,17 @@ float* euclidean_distance_field3d(
       continue;
     }
 
-    // As early as possible, start fetching the
-    // data from RAM b/c the annotated lines below
-    // have 30-50% cache miss.
-    DIJKSTRA_3D_PREFETCH_26WAY(field)
-    DIJKSTRA_3D_PREFETCH_26WAY(dist)
+    if (!queue.empty()) {
+      next_loc = queue.top().value;
+      if (!std::signbit(dist[next_loc])) {
+
+        // As early as possible, start fetching the
+        // data from RAM b/c the annotated lines below
+        // have 30-50% cache miss.
+        DIJKSTRA_3D_PREFETCH_26WAY(field, next_loc)
+        DIJKSTRA_3D_PREFETCH_26WAY(dist, next_loc)
+      }
+    }
 
     if (power_of_two) {
       z = loc >> (xshift + yshift);
@@ -1008,7 +1094,7 @@ float* euclidean_distance_field3d(
       x = loc - sx * (y + z * sy);
     }
 
-    compute_neighborhood(neighborhood, x, y, z, sx, sy, sz);
+    compute_neighborhood(neighborhood, x, y, z, sx, sy, sz, 26, voxel_connectivity_graph);
 
     for (int i = 0; i < NHOOD_SIZE; i++) {
       if (neighborhood[i] == 0) {
